
#include <pthread.h>

typedef enum {
	RESORT_LEVEL_UNKNOWN = 0x00,
	RESORT_LEVEL_TOP     = 0x01,
	RESORT_LEVEL_MID     = 0x02,
	RESORT_LEVEL_BOT     = 0x03,
	RESORT_LEVEL_DEFAULT = RESORT_LEVEL_MID,
} resort_level_t;

typedef struct resort_info_s {
	char *altitude;
	char *header;
	char *info;
	char *summary[2];
	struct {
		char *time;
		char *weather;
		char *wind;
		char *summary;
		char *snow;
		char *rain;
		char *mintemp;
		char *maxtemp;
		char *windchill;
		char *freezelevel;
		char *sunrise;
		char *sunset;
	} day[18];
} resort_info_t;

typedef struct resort_s {
	char *country;
	char *name;
	char *link;
	unsigned int level;
	resort_info_t info;
} resort_t;

typedef struct resort_condition_s {
	int update;
	int success;
	int started;
	int running;
	int stopped;
	pthread_t thread;
	pthread_cond_t cond;
	pthread_mutex_t mutex;
	resort_t *resort;
	char *path;
} resort_condition_t;

int resorts_free (resort_t **resorts);
resort_t ** resorts_list (void);
resort_t ** resorts_read (const char *path);
int resorts_setlevel (const char *path, resort_level_t level);

int resorts_count (resort_t **resorts);
resort_t * resorts_resort (resort_t **resorts, int index);

int resort_insert (const char *path, resort_t *resort);
int resort_remove (const char *path, resort_t *resort);

resort_condition_t * resort_update_start (const char *path, resort_t *resort);
int resort_update_check (resort_condition_t *condition);
int resort_update_finish (resort_condition_t *condition);
