/*
 * Thu Nov 5 01:41:52 EET 2009
 *
 * Alper Akcan <alper.akcan@gmail.com>
 *
 * Parser for snow-forecast print page,
 * I hope they do not change data format :)
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ctype.h>
#include <fcntl.h>
#include <poll.h>

#include <sqlite3.h>

#define DATABASE_VERSION 6
#define DATABASE_PATH "database.db"
#define RESORT_TIMEOUT 20000
#define RESORT_INTERVAL 50
#define RESORT_DATASIZE (256)
#define RESORT_ALLOCSIZE (4096)
#define DATABE_UPDATE_FREQ (60 * 10)

#include "Resort.h"
#include "Resorts.h"

#if !defined(BUILD_UNIX)
#include "Logger.h"
#define debugf(a...) logf(@a);
#else
#define debugf(a...) \
	printf(a); \
	printf(" [%s (%s:%d)]\n", __FUNCTION__, __FILE__, __LINE__);
#endif

#define offsetof_(type, member) ((size_t) &((type *) 0)->member)

typedef struct token_s {
	int order;
	int64_t offset;
} token_t;

#define token_initializer \
	&(token_t) { 0, 0}, &(token_t) { 0, 0}, &(token_t) { 0, 0}, \
	&(token_t) { 0, 0}, &(token_t) { 0, 0}, &(token_t) { 0, 0}, \
	&(token_t) { 0, 0}, &(token_t) { 0, 0}, &(token_t) { 0, 0}, \
	&(token_t) { 0, 0}, &(token_t) { 0, 0}, &(token_t) { 0, 0}

static token_t *tokens[] = {
	&(token_t) { 0, offsetof_(resort_info_t, header) },
	&(token_t) { 1, offsetof_(resort_info_t, info) },
	&(token_t) { 4, offsetof_(resort_info_t, summary[0]) },
	&(token_t) { 5, offsetof_(resort_info_t, summary[1]) },

	&(token_t) { 6, offsetof_(resort_info_t, day[0].time) },
	&(token_t) { 24, offsetof_(resort_info_t, day[0].weather) },
	&(token_t) { 43, offsetof_(resort_info_t, day[0].wind) },
	&(token_t) { 62, offsetof_(resort_info_t, day[0].summary) },
	&(token_t) { 81, offsetof_(resort_info_t, day[0].snow) },
	&(token_t) { 100, offsetof_(resort_info_t, day[0].rain) },
	&(token_t) { 119, offsetof_(resort_info_t, day[0].mintemp) },
	&(token_t) { 138, offsetof_(resort_info_t, day[0].maxtemp) },
	&(token_t) { 157, offsetof_(resort_info_t, day[0].windchill) },
	&(token_t) { 176, offsetof_(resort_info_t, day[0].freezelevel) },
	&(token_t) { 195, offsetof_(resort_info_t, day[0].sunrise) },
	&(token_t) { 214, offsetof_(resort_info_t, day[0].sunset) },

	token_initializer, token_initializer,
	token_initializer, token_initializer, token_initializer,
	token_initializer, token_initializer, token_initializer,
	token_initializer, token_initializer, token_initializer,
	token_initializer, token_initializer, token_initializer,
	token_initializer, token_initializer, token_initializer,

	NULL,
};

typedef struct url_s {
	char *protocol;
	char *user;
	char *password;
	char *host;
	char *port;
	char *path;
	char *filename;
} url_t;

/* url format
 * http://user:password@ip:port/path/to/get
 * http/user:password@ip:port/path/to/get
 */
static inline int url_parse (const char *url, char **protocol, char **user, char **password, char **host, char **port, char **path, char **fname)
{
	char *p;
	char *s;
	char *t;

	char *a;
	char *h;
	char *d;

	t = strdup(url);
	if (t == NULL) {
		goto err;
	}
	if (strncmp(t, "http://", 7) == 0) {
		*protocol = strdup("http");
		if (*protocol == NULL) {
			goto err;
		}
		p = t + 7;
	} else if (strncmp(t, "http/", 5) == 0) {
		*protocol = strdup("http");
		if (*protocol == NULL) {
			goto err;
		}
		p = t + 5;
	} else {
		goto err;
	}

	s = strstr(p, "/");
	if (s == NULL) {
		/* no path found */
		*path = strdup("/");
		*fname = strdup("/");
	} else {
		/* s -> /path/to/get */
		if (strlen(s + 1) == 0) {
			*path = strdup("/");
		} else {
			*path = strdup(s + 1);
		}
		if (*path == NULL) {
			goto err;
		}
		*s = '\0';
		s = strrchr(*path, '/');
		if (s == NULL || strlen(s + 1) == 0) {
			*fname = strdup(*path);
		} else {
			*fname = strdup(s + 1);
		}
		if (*fname == NULL) {
			goto err;
		}
	}
	if (*path == NULL || **path == '\0' ||
	    *fname == NULL || **fname == '\0') {
		goto err;
	}

	h = strstr(p, "@");
	if (h == NULL) {
		/* p -> ip:port */
		h = p;
	} else {
		/* p -> user:password@ip:port */
		*h++ = '\0';
		a = strstr(p, ":");
		if (a == NULL) {
			*user = strdup(p);
			if (*user == NULL) {
				goto err;
			}
		} else {
			*a++ = '\0';
			*user = strdup(p);
			*password = strdup(a);
			if (*user == NULL || *password == NULL) {
				goto err;
			}
		}
	}
	d = strstr(h, ":");
	if (d == NULL) {
		/* p -> ip */
		*host = strdup(h);
		if (*host == NULL) {
			goto err;
		}
	} else {
		/* p -> ip:port */
		*d++ = '\0';
		*host = strdup(h);
		*port = strdup(d);
		if  (*host == NULL || *port == NULL) {
			goto err;
		}
	}
	free(t);
	return 0;
err:	free(*protocol);
	free(*user);
	free(*password);
	free(*host);
	free(*port);
	free(*path);
	free(*fname);
	free(t);
	return -1;
}

static inline url_t * url_init (const char *url)
{
	url_t *u;
	u = (url_t *) malloc(sizeof(url_t));
	if (u == NULL) {
		goto err0;
	}
	memset(u, 0, sizeof(url_t));
	if (url_parse(url, &u->protocol, &u->user, &u->password, &u->host, &u->port, &u->path, &u->filename) != 0) {
		goto err1;
	}
	return u;
err1:	free(u);
err0:	return NULL;
}

static inline int url_uninit (url_t *url)
{
	if (url == NULL) {
		return 0;
	}
	free(url->protocol);
	free(url->user);
	free(url->password);
	free(url->path);
	free(url->port);
	free(url->host);
	free(url->filename);
	free(url);
	return 0;
}

static inline int buffer_dump (resort_info_t *data)
{
	int i;
	printf("%s @ %s m\n", data->header, data->altitude);
	printf("\n");
	printf("info:\n  '%s'\n", data->info);
	printf("\n");
	printf("0-3 days:\n  '%s'\n", data->summary[0]);
	printf("\n");
	printf("4-6 days:\n  '%s'\n", data->summary[1]);
	printf("\n");
	printf("  %15s %18s %9s %13s %6s %6s %9s %9s %11s %13s %9s %9s\n"
		"  %15s %18s %9s %13s %6s %6s %9s %9s %11s %13s %9s %9s\n",
		"time", "weather", "wind", "summary", "snow", "rain", "mintemp", "maxtemp", "windchill", "freezelevel", "sunrise", "sunset",
		"    ",  "      ", "    ", "       ", "  cm", "  cm", "      C", "      C", "        C", "          m", "       ", "      ");
	printf("  %15s %18s %9s %13s %6s %6s %9s %9s %11s %13s %9s %9s\n",
		"-------------", "----------------", "-------", "-----------", "----", "----", "-------", "-------", "---------", "-----------", "-------", "-------");
	for (i = 0; i < 18; i++) {
		printf("  %15s %18s %9s %13s %6s %6s %9s %9s %11s %13s %9s %9s\n",
			data->day[i].time,
			data->day[i].weather,
			data->day[i].wind,
			data->day[i].summary,
			data->day[i].snow,
			data->day[i].rain,
			data->day[i].mintemp,
			data->day[i].maxtemp,
			data->day[i].windchill,
			data->day[i].freezelevel,
			data->day[i].sunrise,
			data->day[i].sunset);
	}
	return 0;
}

static inline int buffer_trim (char *buffer)
{
	char *buf;
	char *ptr;
	char *tmp;
	char **con;
	/*
	 * '\n' -> ' '
	 * '\r' -> ' '
	 * '\t' -> ' '
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (*buf == '\n' || *buf == '\r' || *buf == '\t') {
			*ptr++ = ' ';
			buf++;
		} else if (strncmp(buf, "&nbsp;", 6) == 0) {
			*ptr++ = ' ';
			buf += 6;
		} else if (strncmp(buf, "&deg;", 5) == 0) {
			*ptr++ = ' ';
			buf += 5;
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * '  ' -> ' '
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (*buf == ' ') {
			*ptr++ = *buf++;
			while (*buf == ' ') {
				buf++;
			}
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * ' >' -> '>'
	 * ' <' -> '<'
	 * '< ' -> '<'
	 * '> ' -> '>'
	 * ' =' -> '='
	 * '= ' -> '='
	 */
	buf = buffer;
	ptr = buffer;
	while (buf[0] && buf[1]) {
		if (strncmp(buf, " =", 2) == 0) {
			buf++;
		} else if (strncmp(buf, "= ", 2) == 0) {
			*ptr++ = *buf;
			buf += 2;
		} else if (strncmp(buf, " >", 2) == 0) {
			buf++;
		} else if (strncmp(buf, "> ", 2) == 0) {
			*ptr++ = *buf;
			buf += 2;
		} else if (strncmp(buf, " <", 2) == 0) {
			buf++;
		} else if (strncmp(buf, "< ", 2) == 0) {
			*ptr++ = *buf;
			buf += 2;
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr++ = buf[0];
	*ptr++ = '\0';
	/*
	 * '<b>' -> ''
	 * '</b>' -> ''
	 * '<i>' -> ''
	 * '</i>' -> ''
	 * '<br>' -> ' '
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (strncasecmp(buf, "<b>", 3) == 0) {
			buf += 3;
		} else if (strncasecmp(buf, "</b>", 4) == 0) {
			buf += 4;
		} else if (strncasecmp(buf, "<i>", 3) == 0) {
			buf += 3;
		} else if (strncasecmp(buf, "</i>", 4) == 0) {
			buf += 4;
		} else if (strncasecmp(buf, "<br>", 4) == 0) {
			buf += 4;
			*ptr++ = ' ';
		} else if (strncasecmp(buf, "<br/>", 5) == 0) {
			buf += 5;
			*ptr++ = ' ';
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * '<! *>' -> ''
	 * '<tbody *>' -> ''
	 * '</tbody>' -> ''
	 * '<font *>' -> ''
	 * '</font>' -> ' '
	 * '<div *>' -> ''
	 * '</div>' -> ''
	 * '<hr *>' -> ''
	 * '<html *>' -> '<html>'
	 * '<table *> -> '<table>'
	 * '<td *> -> '<td>'
	 * '<tr *> -> '<tr>'
	 * '<head>*</head> -> ''
	 * ' align="*"' -> ''
	 * ' valign="*"' -> ''
	 * ' width="*"' -> ''
	 * ' height="*"' -> ''
	 * ' bgcolor="*"' -> ''
	 * ' fgcolor="*"' -> ''
	 * '<img src="print_files*"> -> '*'
	 * 'morn- ing' -> 'morning'
	 * 'after- noon' -> 'afternoon'
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (strncmp(buf, "<hr", 3) == 0) {
			buf += 3;
			while (*buf++ != '>') {
				;
			}
		} else if (strncmp(buf, " align=\"", 8) == 0) {
			buf += 8;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, " valign=\"", 9) == 0) {
			buf += 9;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, " width=\"", 8) == 0) {
			buf += 8;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, " height=\"", 9) == 0) {
			buf += 9;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, " bgcolor=\"", 10) == 0) {
			buf += 10;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, " fgcolor=\"", 10) == 0) {
			buf += 10;
			while (*buf++ != '"') {
				;
			}
		} else if (strncmp(buf, "<html", 5) == 0) {
			memcpy(ptr, buf, 5);
			ptr += 5;
			buf += 5;
			while (*buf++ != '>') {
				;
			}
			*ptr++ = '>';
		} else if (strncmp(buf, "<table", 6) == 0) {
			memcpy(ptr, buf, 6);
			ptr += 6;
			buf += 6;
			while (*buf++ != '>') {
				;
			}
			*ptr++ = '>';
		} else if (strncmp(buf, "<td", 3) == 0) {
			memcpy(ptr, buf, 3);
			ptr += 3;
			buf += 3;
			while (*buf++ != '>') {
				;
			}
			*ptr++ = '>';
		} else if (strncmp(buf, "<tr", 3) == 0) {
			memcpy(ptr, buf, 3);
			ptr += 3;
			buf += 3;
			while (*buf++ != '>') {
				;
			}
			*ptr++ = '>';
		} else if (strncmp(buf, "<head>", 6) == 0) {
			buf += 6;
			while (strncmp(buf, "</head>", 7) != 0) {
				buf++;
			}
			buf += 7;
		} else if (strncmp(buf, "<!", 2) == 0) {
			buf += 2;
			while (*buf++ != '>') {
				;
			}
		} else if (strncmp(buf, "<tbody", 6) == 0) {
			buf += 6;
			while (*buf++ != '>') {
				;
			}
		} else if (strncmp(buf, "</tbody>", 8) == 0) {
			buf += 8;
		} else if (strncmp(buf, "<font", 5) == 0) {
			buf += 5;
			while (*buf++ != '>') {
				;
			}
		} else if (strncmp(buf, "</font>", 7) == 0) {
			buf += 7;
			*ptr++ = ' ';
		} else if (strncmp(buf, "<div", 4) == 0) {
			buf += 4;
			while (*buf++ != '>') {
				;
			}
		} else if (strncmp(buf, "</div>", 6) == 0) {
			buf += 6;
#if 1
		} else if (strncmp(buf, "<img src=\"", 10) == 0) {
			buf += 10;
			tmp = strstr(buf, ".gif\">");
			if (tmp != NULL) {
				while (*tmp != '/') {
					tmp--;
				}
				if (tmp >= buf) {
					buf = tmp + 1;
				}
			}
			if (strncmp(buf, "wind", 4) == 0) {
				buf += 4;
				while (strncmp(buf, "metric.gif\">", 12) != 0) {
					*ptr++ = *buf++;
				}
				buf += 12;
			} else {
				while (strncmp(buf, ".gif\">", 6) != 0) {
					*ptr++ = *buf++;
				}
				buf += 6;
			}
#endif
		} else if (strncmp(buf, "after- noon", 11) == 0) {
			memcpy(ptr, "afternoon", 9);
			ptr += 9;
			buf += 11;
		} else if (strncmp(buf, "morn- ing", 9) == 0) {
			memcpy(ptr, "morning", 7);
			ptr += 7;
			buf += 9;
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * '  ' -> ' '
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (*buf == ' ') {
			*ptr++ = *buf++;
			while (*buf == ' ') {
				buf++;
			}
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * ' >' -> '>'
	 * ' <' -> '<'
	 * '< ' -> '<'
	 * '> ' -> '>'
	 */
	buf = buffer;
	ptr = buffer;
	while (buf[0] && buf[1]) {
		if (strncmp(buf, " >", 2) == 0) {
			buf++;
		} else if (strncmp(buf, "> ", 2) == 0) {
			*ptr++ = *buf;
			buf += 2;
		} else if (strncmp(buf, " <", 2) == 0) {
			buf++;
		} else if (strncmp(buf, "< ", 2) == 0) {
			*ptr++ = *buf;
			buf += 2;
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr++ = buf[0];
	*ptr++ = '\0';
	/*
	 * tokenize with '\a'
	 */
	buf = buffer;
	ptr = buffer;
	char *converts[] = {
		"<td>", "</td>",
		"<tr>", "</tr>",
		"<table>", "</table>",
		"<p>", "</p>",
		"<html>", "</html>",
		"<body>", "</body>",
		NULL,
	};
	while (*buf) {
		for (con = converts; *con; con++) {
			if (strncmp(buf, *con, strlen(*con)) == 0) {
				buf += strlen(*con);
				*ptr++ = '\a';
				break;
			}
		}
		if (*con == NULL) {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	/*
	 * '\a\a' -> '\a'
	 */
	buf = buffer;
	ptr = buffer;
	while (*buf) {
		if (*buf == '\a') {
			*ptr++ = *buf++;
			while (*buf == '\a') {
				buf++;
			}
		} else {
			*ptr++ = *buf++;
		}
	}
	*ptr = '\0';
	return 0;
}

static inline resort_info_t * buffer_parse (char *buffer)
{
	int order;
	char *buf;
	char *ptr;
	char *tmp;
	resort_info_t *data;
	token_t **tok;
	token_t **tokp;
	if (buffer_trim(buffer) != 0) {
		return NULL;
	}
	data = (resort_info_t *) malloc(sizeof(resort_info_t));
	if (data == NULL) {
		return NULL;
	}
	memset(data, 0, sizeof(resort_info_t));
	for (order = 1, tok = &tokens[16], tokp = &tokens[4]; *tok; tok += 12, tokp += 12, order++) {
		(*tok[0]).offset = offsetof_(resort_info_t, day[order].time);
		(*tok[0]).order = (*tokp[0]).order + 1;
		(*tok[1]).offset = offsetof_(resort_info_t, day[order].weather);
		(*tok[1]).order = (*tokp[1]).order + 1;
		(*tok[2]).offset = offsetof_(resort_info_t, day[order].wind);
		(*tok[2]).order = (*tokp[2]).order + 1;
		(*tok[3]).offset = offsetof_(resort_info_t, day[order].summary);
		(*tok[3]).order = (*tokp[3]).order + 1;
		(*tok[4]).offset = offsetof_(resort_info_t, day[order].snow);
		(*tok[4]).order = (*tokp[4]).order + 1;
		(*tok[5]).offset = offsetof_(resort_info_t, day[order].rain);
		(*tok[5]).order = (*tokp[5]).order + 1;
		(*tok[6]).offset = offsetof_(resort_info_t, day[order].mintemp);
		(*tok[6]).order = (*tokp[6]).order + 1;
		(*tok[7]).offset = offsetof_(resort_info_t, day[order].maxtemp);
		(*tok[7]).order = (*tokp[7]).order + 1;
		(*tok[8]).offset = offsetof_(resort_info_t, day[order].windchill);
		(*tok[8]).order = (*tokp[8]).order + 1;
		(*tok[9]).offset = offsetof_(resort_info_t, day[order].freezelevel);
		(*tok[9]).order = (*tokp[9]).order + 1;
		(*tok[10]).offset = offsetof_(resort_info_t, day[order].sunrise);
		(*tok[10]).order = (*tokp[10]).order + 1;
		(*tok[11]).offset = offsetof_(resort_info_t, day[order].sunset);
		(*tok[11]).order = (*tokp[11]).order + 1;
	}
	order = 0;
	buf = buffer + 1;
	while (*buf) {
		ptr = buf;
		while (*buf != '\a') {
			buf++;
		}
		if (buf > ptr) {
			*buf++ = '\0';
			//printf("order[%d]: '%s'\n", order, ptr);
			for (tok = tokens; *tok; tok++) {
				if ((*tok)->order == order) {
					*(char **) (((char *) data) + (*tok)->offset) = ptr;
				}
			}
			order++;
		} else {
			buf++;
		}
	}
	if (data->header != NULL) {
		ptr = strcasestr(data->header, " Weather Forecast");
		if (ptr != NULL) {
			*ptr = '\0';
		}
	}
	if (data->info != NULL) {
		ptr = data->info;
		while (*ptr && *ptr != '(') {
			ptr++;
		}
		ptr++;
		for (buf = ptr; *buf && *buf != ' '; buf++) {
			;
		}
		tmp = (char *) malloc(buf - ptr + 1);
		if (tmp != NULL) {
			memcpy(tmp, ptr, buf - ptr);
			tmp[buf - ptr] = '\0';
			data->altitude = tmp;
		}
	}
	for (order = 0; order < 18; order++) {
		data->day[order].time[4] = toupper(data->day[order].time[4]);
		//data->day[order].time[5] = '\0';
	}
	return data;
}

static inline char * buffer_read_file (char *path)
{
	FILE *fi;
	char *buffer;
	struct stat stbuf;
	if (stat(path + 7, &stbuf) != 0) {
		return NULL;
	}
	if (stbuf.st_size <= 0) {
		return NULL;
	}
	buffer = (char *) malloc(stbuf.st_size + 1);
	if (buffer == NULL) {
		return NULL;
	}
	memset(buffer, 0, stbuf.st_size + 1);
	fi = fopen(path + 7, "r+");
	if (fi == NULL) {
		free(buffer);
		return NULL;
	}
	fread(buffer, 1, stbuf.st_size, fi);
	fclose(fi);
	return buffer;
}

static inline int buffer_check_status (int sock, int event, int timeout, int interval, resort_condition_t *condition)
{
	int timeval;
	struct pollfd pfd;
	timeval = 0;
	while (timeval < timeout) {
		if (condition != NULL) {
			pthread_mutex_lock(&condition->mutex);
			if (condition->running == 0) {
				pthread_mutex_unlock(&condition->mutex);
				break;
			}
			pthread_mutex_unlock(&condition->mutex);
		}
		memset(&pfd, 0, sizeof(struct pollfd));
		pfd.fd = sock;
		pfd.events = event;
		if (poll(&pfd, 1, interval) > 0) {
			if ((pfd.revents & event) == event) {
				debugf("success");
				return 0;
			}
		}
		timeval += interval;
	}
	debugf("error");
	return -1;
}

static inline char * buffer_read_http (char *path, resort_condition_t *condition)
{
	int ret;
	int len;
	int sock;
	int size;
	int code;
	long flags;
	url_t *url;
	struct addrinfo hints;
	struct addrinfo *result;
	struct addrinfo *resultp;
	char *tmp;
	char *buffer;
	char *buffert;
	const char *format =
		"GET /%s HTTP/1.0\r\n"
		"User-Agent: Simple HTTP 1.0\r\n"
		"Accept: */*\r\n"
		"Host: %s\r\n"
		"Connection: Close\r\n"
		"\r\n";

	sock = -1;
	buffer = NULL;
	result = NULL;

	url = url_init(path);
	if (url == NULL) {
		debugf("url_init(%s) failed", path);
		goto out;
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	ret = getaddrinfo(url->host, (url->port) ? url->port : "80", &hints, &result);
	if (ret != 0) {
		debugf("getaddrinfo() failed");
		goto out;
	}
	for (resultp = result; resultp != NULL; resultp = resultp->ai_next) {
		sock = socket(resultp->ai_family, resultp->ai_socktype, resultp->ai_protocol);
		if (sock < 0) {
			continue;
		}
		flags = fcntl(sock, F_GETFL);
		if (flags < 0) {
			close(sock);
			sock = -1;
			continue;
		}
		flags |= O_NONBLOCK;
		if (fcntl(sock, F_SETFL, flags) < 0) {
			close(sock);
			sock = -1;
			continue;
		}
		ret = connect(sock, resultp->ai_addr, resultp->ai_addrlen);
		if (ret == -1) {
			ret = buffer_check_status(sock, POLLOUT, RESORT_TIMEOUT, RESORT_INTERVAL, condition);
			if (ret < 0) {
				close(sock);
				sock = -1;
				continue;
			}
		}
		flags &= ~O_NONBLOCK;
		if (fcntl(sock, F_SETFL, flags) < 0) {
			close(sock);
			sock = -1;
			continue;
		}
		break;
	}
	if (resultp == NULL) {
		debugf("connect() failed");
		goto out;
	}

	ret = asprintf(&buffer, format, url->path, url->host);
	if (ret <= 0) {
		goto out;
	}
	len = 0;
	size = strlen(buffer);
	while (len < size) {
		ret = buffer_check_status(sock, POLLOUT, RESORT_TIMEOUT, RESORT_INTERVAL, condition);
		if (ret < 0) {
			goto out;
		}
		code = (RESORT_DATASIZE < (size - len)) ? RESORT_DATASIZE : (size - len);
		ret = write(sock, buffer + len,  code);
		if (ret <= 0) {
			goto out;
		}
		len += ret;
	}
	free(buffer);

	buffer = NULL;
	buffert = NULL;
	len = 0;
	size = 0;
	while (1) {
		if (len + RESORT_DATASIZE > size) {
			buffert = (char *) malloc(len + RESORT_ALLOCSIZE + 1);
			if (buffert == NULL) {
				goto out;
			}
			size = len + RESORT_ALLOCSIZE;
			if (len > 0) {
				memcpy(buffert, buffer, len);
			}
			memset(buffert + len, 0, size - len + 1);
			free(buffer);
			buffer = buffert;
		}
		ret = buffer_check_status(sock, POLLIN, RESORT_TIMEOUT, RESORT_INTERVAL, condition);
		if (ret < 0) {
			break;
		}
		ret = read(sock, buffer + len, RESORT_DATASIZE);
		if (ret <= 0) {
			break;
		}
		len += ret;
	}
	buffert = buffer;
	while (!isspace(*buffert) && *buffert != '\0') {
		buffert++;
	}
	while (isspace(*buffert)) {
		buffert++;
	}
	code = strtol(buffert, NULL, 10);
	if (code != 200) {
		goto out;
	}
	tmp = strcasestr(buffert, "Content-Length");
	if (tmp == NULL) {
		debugf("could not find content-length");
		goto out;
	}
	while (*tmp != '\0' && *tmp != ':') {
		tmp++;
	}
	if (*tmp != ':') {
		debugf("could not find content-length");
		goto out;
	}
	*tmp++ = '\0';
	while (isspace(*tmp)) {
		tmp++;
	}
	size = atoi(tmp);
	buffert = strstr(tmp, "\r\n\r\n");
	if (buffert == NULL) {
		debugf("could not find content start tag");
		goto out;
	} else {
		buffert += 4;
		tmp = (char *) malloc(strlen(buffert) + 1);
		if (tmp == NULL) {
			goto out;
		}
		memcpy(tmp, buffert, strlen(buffert) + 1);
		free(buffer);
		buffer = tmp;
	}
	if (size != strlen(buffer)) {
		debugf("content size mismatch (%d != %llu)", size, strlen(buffer));
		goto out;
	}
	close(sock);
	url_uninit(url);
	freeaddrinfo(result);
	return buffer;
out:
	if (buffer != NULL) {
		free(buffer);
	}
	if (sock >= 0) {
		close(sock);
	}
	if (url != NULL) {
		url_uninit(url);
	}
	if (result != NULL) {
		freeaddrinfo(result);
	}
	return NULL;
}

static inline char * buffer_read (char *path, resort_condition_t *condition)
{
	char *buffer;
	debugf("fetching from: '%s'", path);
	if (strncmp(path, "resort://", 9) == 0) {
		int count;
		int rcount;
		char *link;
		char *alti;
		const resort_t **res;
		const resort_t **resp;
		path += 9;
		alti = "mid";
		if (strncmp(path, "top@", 4) == 0) {
			path += 4;
			alti = "top";
		} else if (strncmp(path, "mid@", 4) == 0) {
			path += 4;
			alti = "mid";
		} else if (strncmp(path, "bot@", 4) == 0) {
			path += 4;
			alti = "bot";
		}
		resp = NULL;
		count = 0;
		rcount = 0;
		for (res = resorts_table; *res; res++) {
			if (strcasestr((*res)->link, path) != NULL) {
				count++;
				resp = res;
				printf("resort: '%s' (%s,%s)\n", (*res)->link, (*res)->name, (*res)->country);
			}
			rcount++;
		}
		if (count == 1 && resp != NULL) {
			link = (char *) malloc(strlen("http://www.snow-forecast.com/resorts") + 1 + strlen((*resp)->link) + 1 + strlen("6day/top/print") + 1);
			if (link != NULL) {
				sprintf(link, "http://www.snow-forecast.com/resorts/%s/6day/%s/print", (*resp)->link, alti);
				buffer = buffer_read(link, condition);
				free(link);
				return buffer;
			}
		}
	} else if (strncmp(path, "file://", 7) == 0) {
		return buffer_read_file(path);
	} else if (strncmp(path, "http://", 7) == 0) {
		return buffer_read_http(path, condition);
	}
	return NULL;
}

static inline sqlite3 * database_create (const char *path)
{
	int res;
	char *query;
	sqlite3 *db;
	unlink(path);
	debugf("enter (path:%s)", path);
	res = sqlite3_open(path, &db);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	query = sqlite3_mprintf("pragma user_version=%d;", DATABASE_VERSION);
	if (query == NULL) {
		sqlite3_close(db);
		return NULL;
	}
	res = sqlite3_exec(db, query, 0, 0, 0);
	sqlite3_free(query);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	res = sqlite3_exec(db,
		"CREATE TABLE RESORT ("
		"  ID INTEGER PRIMARY KEY AUTOINCREMENT,"
		"  TIME INTEGER NOT NULL,"
		"  NAME TEXT NOT NULL,"
		"  COUNTRY TEXT NOT NULL,"
		"  LINK TEXT NOT NULL,"
		"  TMB INTEGER NOT NULL,"
		"  ALTITUDE TEXT NOT NULL,"
		"  INFO TEXT NOT NULL,"
		"  SUMMARY0 TEXT NOT NULL,"
		"  SUMMARY1 TEXT NOT NULL,"
		"  DAY0TIME TEXT NOT NULL, DAY0SUMMARY TEXT NOT NULL, DAY0WIND TEXT NOT NULL, DAY0SNOW TEXT NOT NULL, DAY0RAIN TEXT NOT NULL, DAY0TEMPMIN TEXT NOT NULL, DAY0TEMPMAX TEXT NOT NULL, DAY0TEMPCHILL TEXT NOT NULL, DAY0FREEZE TEXT NOT NULL,"
		"  DAY1TIME TEXT NOT NULL, DAY1SUMMARY TEXT NOT NULL, DAY1WIND TEXT NOT NULL, DAY1SNOW TEXT NOT NULL, DAY1RAIN TEXT NOT NULL, DAY1TEMPMIN TEXT NOT NULL, DAY1TEMPMAX TEXT NOT NULL, DAY1TEMPCHILL TEXT NOT NULL, DAY1FREEZE TEXT NOT NULL,"
		"  DAY2TIME TEXT NOT NULL, DAY2SUMMARY TEXT NOT NULL, DAY2WIND TEXT NOT NULL, DAY2SNOW TEXT NOT NULL, DAY2RAIN TEXT NOT NULL, DAY2TEMPMIN TEXT NOT NULL, DAY2TEMPMAX TEXT NOT NULL, DAY2TEMPCHILL TEXT NOT NULL, DAY2FREEZE TEXT NOT NULL,"
		"  DAY3TIME TEXT NOT NULL, DAY3SUMMARY TEXT NOT NULL, DAY3WIND TEXT NOT NULL, DAY3SNOW TEXT NOT NULL, DAY3RAIN TEXT NOT NULL, DAY3TEMPMIN TEXT NOT NULL, DAY3TEMPMAX TEXT NOT NULL, DAY3TEMPCHILL TEXT NOT NULL, DAY3FREEZE TEXT NOT NULL,"
		"  DAY4TIME TEXT NOT NULL, DAY4SUMMARY TEXT NOT NULL, DAY4WIND TEXT NOT NULL, DAY4SNOW TEXT NOT NULL, DAY4RAIN TEXT NOT NULL, DAY4TEMPMIN TEXT NOT NULL, DAY4TEMPMAX TEXT NOT NULL, DAY4TEMPCHILL TEXT NOT NULL, DAY4FREEZE TEXT NOT NULL,"
		"  DAY5TIME TEXT NOT NULL, DAY5SUMMARY TEXT NOT NULL, DAY5WIND TEXT NOT NULL, DAY5SNOW TEXT NOT NULL, DAY5RAIN TEXT NOT NULL, DAY5TEMPMIN TEXT NOT NULL, DAY5TEMPMAX TEXT NOT NULL, DAY5TEMPCHILL TEXT NOT NULL, DAY5FREEZE TEXT NOT NULL,"
		"  DAY6TIME TEXT NOT NULL, DAY6SUMMARY TEXT NOT NULL, DAY6WIND TEXT NOT NULL, DAY6SNOW TEXT NOT NULL, DAY6RAIN TEXT NOT NULL, DAY6TEMPMIN TEXT NOT NULL, DAY6TEMPMAX TEXT NOT NULL, DAY6TEMPCHILL TEXT NOT NULL, DAY6FREEZE TEXT NOT NULL,"
		"  DAY7TIME TEXT NOT NULL, DAY7SUMMARY TEXT NOT NULL, DAY7WIND TEXT NOT NULL, DAY7SNOW TEXT NOT NULL, DAY7RAIN TEXT NOT NULL, DAY7TEMPMIN TEXT NOT NULL, DAY7TEMPMAX TEXT NOT NULL, DAY7TEMPCHILL TEXT NOT NULL, DAY7FREEZE TEXT NOT NULL,"
		"  DAY8TIME TEXT NOT NULL, DAY8SUMMARY TEXT NOT NULL, DAY8WIND TEXT NOT NULL, DAY8SNOW TEXT NOT NULL, DAY8RAIN TEXT NOT NULL, DAY8TEMPMIN TEXT NOT NULL, DAY8TEMPMAX TEXT NOT NULL, DAY8TEMPCHILL TEXT NOT NULL, DAY8FREEZE TEXT NOT NULL,"
		"  DAY9TIME TEXT NOT NULL, DAY9SUMMARY TEXT NOT NULL, DAY9WIND TEXT NOT NULL, DAY9SNOW TEXT NOT NULL, DAY9RAIN TEXT NOT NULL, DAY9TEMPMIN TEXT NOT NULL, DAY9TEMPMAX TEXT NOT NULL, DAY9TEMPCHILL TEXT NOT NULL, DAY9FREEZE TEXT NOT NULL,"
		"  DAY10TIME TEXT NOT NULL, DAY10SUMMARY TEXT NOT NULL, DAY10WIND TEXT NOT NULL, DAY10SNOW TEXT NOT NULL, DAY10RAIN TEXT NOT NULL, DAY10TEMPMIN TEXT NOT NULL, DAY10TEMPMAX TEXT NOT NULL, DAY10TEMPCHILL TEXT NOT NULL, DAY10FREEZE TEXT NOT NULL,"
		"  DAY11TIME TEXT NOT NULL, DAY11SUMMARY TEXT NOT NULL, DAY11WIND TEXT NOT NULL, DAY11SNOW TEXT NOT NULL, DAY11RAIN TEXT NOT NULL, DAY11TEMPMIN TEXT NOT NULL, DAY11TEMPMAX TEXT NOT NULL, DAY11TEMPCHILL TEXT NOT NULL, DAY11FREEZE TEXT NOT NULL,"
		"  DAY12TIME TEXT NOT NULL, DAY12SUMMARY TEXT NOT NULL, DAY12WIND TEXT NOT NULL, DAY12SNOW TEXT NOT NULL, DAY12RAIN TEXT NOT NULL, DAY12TEMPMIN TEXT NOT NULL, DAY12TEMPMAX TEXT NOT NULL, DAY12TEMPCHILL TEXT NOT NULL, DAY12FREEZE TEXT NOT NULL,"
		"  DAY13TIME TEXT NOT NULL, DAY13SUMMARY TEXT NOT NULL, DAY13WIND TEXT NOT NULL, DAY13SNOW TEXT NOT NULL, DAY13RAIN TEXT NOT NULL, DAY13TEMPMIN TEXT NOT NULL, DAY13TEMPMAX TEXT NOT NULL, DAY13TEMPCHILL TEXT NOT NULL, DAY13FREEZE TEXT NOT NULL,"
		"  DAY14TIME TEXT NOT NULL, DAY14SUMMARY TEXT NOT NULL, DAY14WIND TEXT NOT NULL, DAY14SNOW TEXT NOT NULL, DAY14RAIN TEXT NOT NULL, DAY14TEMPMIN TEXT NOT NULL, DAY14TEMPMAX TEXT NOT NULL, DAY14TEMPCHILL TEXT NOT NULL, DAY14FREEZE TEXT NOT NULL,"
		"  DAY15TIME TEXT NOT NULL, DAY15SUMMARY TEXT NOT NULL, DAY15WIND TEXT NOT NULL, DAY15SNOW TEXT NOT NULL, DAY15RAIN TEXT NOT NULL, DAY15TEMPMIN TEXT NOT NULL, DAY15TEMPMAX TEXT NOT NULL, DAY15TEMPCHILL TEXT NOT NULL, DAY15FREEZE TEXT NOT NULL,"
		"  DAY16TIME TEXT NOT NULL, DAY16SUMMARY TEXT NOT NULL, DAY16WIND TEXT NOT NULL, DAY16SNOW TEXT NOT NULL, DAY16RAIN TEXT NOT NULL, DAY16TEMPMIN TEXT NOT NULL, DAY16TEMPMAX TEXT NOT NULL, DAY16TEMPCHILL TEXT NOT NULL, DAY16FREEZE TEXT NOT NULL,"
		"  DAY17TIME TEXT NOT NULL, DAY17SUMMARY TEXT NOT NULL, DAY17WIND TEXT NOT NULL, DAY17SNOW TEXT NOT NULL, DAY17RAIN TEXT NOT NULL, DAY17TEMPMIN TEXT NOT NULL, DAY17TEMPMAX TEXT NOT NULL, DAY17TEMPCHILL TEXT NOT NULL, DAY17FREEZE TEXT NOT NULL,"
		"  UNIQUE(LINK));",
		0, 0, 0);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	res = sqlite3_exec(db, "CREATE INDEX INDEX_RESORT_ID on RESORT(ID);", 0, 0, 0);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	res = sqlite3_exec(db, "CREATE INDEX INDEX_RESORT_NAME on RESORT(NAME);", 0, 0, 0);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	res = sqlite3_exec(db, "CREATE INDEX INDEX_RESORT_LINK on RESORT(NAME, LINK);", 0, 0, 0);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return NULL;
	}
	debugf("leave (db:%p)", db);
	return db;
}

static inline sqlite3 * database_open (const char *path)
{
	int res;
	int rows;
	sqlite3 *db;
	char **results;
	debugf("enter (path:%s)", path);
	res = sqlite3_open(path, &db);
	if (res != SQLITE_OK) {
		sqlite3_close(db);
		return database_create(path);
	}
	res = sqlite3_get_table(db, "pragma user_version", &results, &rows, NULL, NULL);
	if (res != SQLITE_OK) {
		sqlite3_free_table(results);
		sqlite3_close(db);
		return database_create(path);
	}
	if (rows <= 0 || atoi(results[1]) != DATABASE_VERSION) {
		sqlite3_free_table(results);
		sqlite3_close(db);
		return database_create(path);
	}
	sqlite3_free_table(results);
	debugf("leave (db:%p)", db);
	return db;
}

static inline int database_close (sqlite3 *database)
{
	return sqlite3_close(database);
}

int resort_insert (const char *path, resort_t *resort)
{
	int res;
	char *query;
	sqlite3 *db;
	debugf("enter");
	db = database_open(path);
	if (db == NULL) {
		debugf("error");
		return -1;
	}
	query = sqlite3_mprintf(
		"INSERT into RESORT"
		"  (TIME, NAME, COUNTRY, LINK, TMB,"
		"   ALTITUDE, INFO, SUMMARY0, SUMMARY1,"
		"   DAY0TIME, DAY0SUMMARY, DAY0WIND, DAY0SNOW, DAY0RAIN, DAY0TEMPMIN, DAY0TEMPMAX, DAY0TEMPCHILL, DAY0FREEZE,"
		"   DAY1TIME, DAY1SUMMARY, DAY1WIND, DAY1SNOW, DAY1RAIN, DAY1TEMPMIN, DAY1TEMPMAX, DAY1TEMPCHILL, DAY1FREEZE,"
		"   DAY2TIME, DAY2SUMMARY, DAY2WIND, DAY2SNOW, DAY2RAIN, DAY2TEMPMIN, DAY2TEMPMAX, DAY2TEMPCHILL, DAY2FREEZE,"
		"   DAY3TIME, DAY3SUMMARY, DAY3WIND, DAY3SNOW, DAY3RAIN, DAY3TEMPMIN, DAY3TEMPMAX, DAY3TEMPCHILL, DAY3FREEZE,"
		"   DAY4TIME, DAY4SUMMARY, DAY4WIND, DAY4SNOW, DAY4RAIN, DAY4TEMPMIN, DAY4TEMPMAX, DAY4TEMPCHILL, DAY4FREEZE,"
		"   DAY5TIME, DAY5SUMMARY, DAY5WIND, DAY5SNOW, DAY5RAIN, DAY5TEMPMIN, DAY5TEMPMAX, DAY5TEMPCHILL, DAY5FREEZE,"
		"   DAY6TIME, DAY6SUMMARY, DAY6WIND, DAY6SNOW, DAY6RAIN, DAY6TEMPMIN, DAY6TEMPMAX, DAY6TEMPCHILL, DAY6FREEZE,"
		"   DAY7TIME, DAY7SUMMARY, DAY7WIND, DAY7SNOW, DAY7RAIN, DAY7TEMPMIN, DAY7TEMPMAX, DAY7TEMPCHILL, DAY7FREEZE,"
		"   DAY8TIME, DAY8SUMMARY, DAY8WIND, DAY8SNOW, DAY8RAIN, DAY8TEMPMIN, DAY8TEMPMAX, DAY8TEMPCHILL, DAY8FREEZE,"
		"   DAY9TIME, DAY9SUMMARY, DAY9WIND, DAY9SNOW, DAY9RAIN, DAY9TEMPMIN, DAY9TEMPMAX, DAY9TEMPCHILL, DAY9FREEZE,"
		"   DAY10TIME, DAY10SUMMARY, DAY10WIND, DAY10SNOW, DAY10RAIN, DAY10TEMPMIN, DAY10TEMPMAX, DAY10TEMPCHILL, DAY10FREEZE,"
		"   DAY11TIME, DAY11SUMMARY, DAY11WIND, DAY11SNOW, DAY11RAIN, DAY11TEMPMIN, DAY11TEMPMAX, DAY11TEMPCHILL, DAY11FREEZE,"
		"   DAY12TIME, DAY12SUMMARY, DAY12WIND, DAY12SNOW, DAY12RAIN, DAY12TEMPMIN, DAY12TEMPMAX, DAY12TEMPCHILL, DAY12FREEZE,"
		"   DAY13TIME, DAY13SUMMARY, DAY13WIND, DAY13SNOW, DAY13RAIN, DAY13TEMPMIN, DAY13TEMPMAX, DAY13TEMPCHILL, DAY13FREEZE,"
		"   DAY14TIME, DAY14SUMMARY, DAY14WIND, DAY14SNOW, DAY14RAIN, DAY14TEMPMIN, DAY14TEMPMAX, DAY14TEMPCHILL, DAY14FREEZE,"
		"   DAY15TIME, DAY15SUMMARY, DAY15WIND, DAY15SNOW, DAY15RAIN, DAY15TEMPMIN, DAY15TEMPMAX, DAY15TEMPCHILL, DAY15FREEZE,"
		"   DAY16TIME, DAY16SUMMARY, DAY16WIND, DAY16SNOW, DAY16RAIN, DAY16TEMPMIN, DAY16TEMPMAX, DAY16TEMPCHILL, DAY16FREEZE,"
		"   DAY17TIME, DAY17SUMMARY, DAY17WIND, DAY17SNOW, DAY17RAIN, DAY17TEMPMIN, DAY17TEMPMAX, DAY17TEMPCHILL, DAY17FREEZE"
		"  )"
		"  values"
		"  ('0', '%s', '%s', '%s', '%u',"
		"   'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a',"
		"   'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a', 'n/a'"
		"  );",
		resort->name,
		resort->country,
		resort->link,
		resort->level);
	res = sqlite3_exec(db, query, 0, 0, 0);
	sqlite3_free(query);
	database_close(db);
	if (res == SQLITE_OK) {
		debugf("leave");
		return 0;
	} else {
		debugf("error");
		return -1;
	}
}

int resort_remove (const char *path, resort_t *resort)
{
	int res;
	char *query;
	sqlite3 *db;
	db = database_open(path);
	if (db == NULL) {
		return -1;
	}
	query = sqlite3_mprintf(
		"DELETE from RESORT"
		"  where"
		"  LINK='%s';",
		resort->link);
	res = sqlite3_exec(db, query, 0, 0, 0);
	sqlite3_free(query);
	database_close(db);
	if (res == SQLITE_OK) {
		return 0;
	} else {
		return -1;
	}
}

static int resort_update_time (const char *path, resort_t *resort)
{
	int t = 0;
	sqlite3 *db;
	char *query;
	const char *format =
		" SELECT"
		"  r.TIME from RESORT r"
		" where r.LINK='%s';";
	sqlite3_stmt *stmt;
	debugf("enter");
	query = sqlite3_mprintf(format, resort->link);
	if (query == NULL) {
		debugf("error");
		return 0;
	}
	db = database_open(path);
	if (db == NULL) {
		sqlite3_free(query);
		debugf("error");
		return 0;
	}
	if (sqlite3_prepare_v2(db, query, -1, &stmt, NULL) == SQLITE_OK) {
		while (sqlite3_step(stmt) == SQLITE_ROW) {
			t = atoi((const char *) sqlite3_column_text(stmt, 0));
		}
	}
	debugf("got update time:%d", t);
	database_close(db);
	return t;
}

static int resort_update (const char *path, const char *link, resort_info_t *info, resort_level_t level)
{
	int r;
	time_t t;
	sqlite3 *db;
	char *query;
	const char *format =
		"UPDATE RESORT set"
		"  TIME='%u', TMB='%u', ALTITUDE='%s', INFO='%s', SUMMARY0='%s', SUMMARY1='%s',"
		"  DAY0TIME='%s', DAY0SUMMARY='%s', DAY0WIND='%s', DAY0SNOW='%s', DAY0RAIN='%s', DAY0TEMPMIN='%s', DAY0TEMPMAX='%s', DAY0TEMPCHILL='%s', DAY0FREEZE='%s',"
		"  DAY1TIME='%s', DAY1SUMMARY='%s', DAY1WIND='%s', DAY1SNOW='%s', DAY1RAIN='%s', DAY1TEMPMIN='%s', DAY1TEMPMAX='%s', DAY1TEMPCHILL='%s', DAY1FREEZE='%s',"
		"  DAY2TIME='%s', DAY2SUMMARY='%s', DAY2WIND='%s', DAY2SNOW='%s', DAY2RAIN='%s', DAY2TEMPMIN='%s', DAY2TEMPMAX='%s', DAY2TEMPCHILL='%s', DAY2FREEZE='%s',"
		"  DAY3TIME='%s', DAY3SUMMARY='%s', DAY3WIND='%s', DAY3SNOW='%s', DAY3RAIN='%s', DAY3TEMPMIN='%s', DAY3TEMPMAX='%s', DAY3TEMPCHILL='%s', DAY3FREEZE='%s',"
		"  DAY4TIME='%s', DAY4SUMMARY='%s', DAY4WIND='%s', DAY4SNOW='%s', DAY4RAIN='%s', DAY4TEMPMIN='%s', DAY4TEMPMAX='%s', DAY4TEMPCHILL='%s', DAY4FREEZE='%s',"
		"  DAY5TIME='%s', DAY5SUMMARY='%s', DAY5WIND='%s', DAY5SNOW='%s', DAY5RAIN='%s', DAY5TEMPMIN='%s', DAY5TEMPMAX='%s', DAY5TEMPCHILL='%s', DAY5FREEZE='%s',"
		"  DAY6TIME='%s', DAY6SUMMARY='%s', DAY6WIND='%s', DAY6SNOW='%s', DAY6RAIN='%s', DAY6TEMPMIN='%s', DAY6TEMPMAX='%s', DAY6TEMPCHILL='%s', DAY6FREEZE='%s',"
		"  DAY7TIME='%s', DAY7SUMMARY='%s', DAY7WIND='%s', DAY7SNOW='%s', DAY7RAIN='%s', DAY7TEMPMIN='%s', DAY7TEMPMAX='%s', DAY7TEMPCHILL='%s', DAY7FREEZE='%s',"
		"  DAY8TIME='%s', DAY8SUMMARY='%s', DAY8WIND='%s', DAY8SNOW='%s', DAY8RAIN='%s', DAY8TEMPMIN='%s', DAY8TEMPMAX='%s', DAY8TEMPCHILL='%s', DAY8FREEZE='%s',"
		"  DAY9TIME='%s', DAY9SUMMARY='%s', DAY9WIND='%s', DAY9SNOW='%s', DAY9RAIN='%s', DAY9TEMPMIN='%s', DAY9TEMPMAX='%s', DAY9TEMPCHILL='%s', DAY9FREEZE='%s',"
		"  DAY10TIME='%s', DAY10SUMMARY='%s', DAY10WIND='%s', DAY10SNOW='%s', DAY10RAIN='%s', DAY10TEMPMIN='%s', DAY10TEMPMAX='%s', DAY10TEMPCHILL='%s', DAY10FREEZE='%s',"
		"  DAY11TIME='%s', DAY11SUMMARY='%s', DAY11WIND='%s', DAY11SNOW='%s', DAY11RAIN='%s', DAY11TEMPMIN='%s', DAY11TEMPMAX='%s', DAY11TEMPCHILL='%s', DAY11FREEZE='%s',"
		"  DAY12TIME='%s', DAY12SUMMARY='%s', DAY12WIND='%s', DAY12SNOW='%s', DAY12RAIN='%s', DAY12TEMPMIN='%s', DAY12TEMPMAX='%s', DAY12TEMPCHILL='%s', DAY12FREEZE='%s',"
		"  DAY13TIME='%s', DAY13SUMMARY='%s', DAY13WIND='%s', DAY13SNOW='%s', DAY13RAIN='%s', DAY13TEMPMIN='%s', DAY13TEMPMAX='%s', DAY13TEMPCHILL='%s', DAY13FREEZE='%s',"
		"  DAY14TIME='%s', DAY14SUMMARY='%s', DAY14WIND='%s', DAY14SNOW='%s', DAY14RAIN='%s', DAY14TEMPMIN='%s', DAY14TEMPMAX='%s', DAY14TEMPCHILL='%s', DAY14FREEZE='%s',"
		"  DAY15TIME='%s', DAY15SUMMARY='%s', DAY15WIND='%s', DAY15SNOW='%s', DAY15RAIN='%s', DAY15TEMPMIN='%s', DAY15TEMPMAX='%s', DAY15TEMPCHILL='%s', DAY15FREEZE='%s',"
		"  DAY16TIME='%s', DAY16SUMMARY='%s', DAY16WIND='%s', DAY16SNOW='%s', DAY16RAIN='%s', DAY16TEMPMIN='%s', DAY16TEMPMAX='%s', DAY16TEMPCHILL='%s', DAY16FREEZE='%s',"
		"  DAY17TIME='%s', DAY17SUMMARY='%s', DAY17WIND='%s', DAY17SNOW='%s', DAY17RAIN='%s', DAY17TEMPMIN='%s', DAY17TEMPMAX='%s', DAY17TEMPCHILL='%s', DAY17FREEZE='%s'"
		"  where LINK='%s';";
	t = time(NULL);
	db = database_open(path);
	if (db == NULL) {
		return -1;
	}
	query = sqlite3_mprintf(format,
				t, level, info->altitude, info->info, info->summary[0], info->summary[1],
				info->day[0].time, info->day[0].summary, info->day[0].wind, info->day[0].snow, info->day[0].rain, info->day[0].mintemp, info->day[0].maxtemp, info->day[0].windchill, info->day[0].freezelevel,
				info->day[1].time, info->day[1].summary, info->day[1].wind, info->day[1].snow, info->day[1].rain, info->day[1].mintemp, info->day[1].maxtemp, info->day[1].windchill, info->day[1].freezelevel,
				info->day[2].time, info->day[2].summary, info->day[2].wind, info->day[2].snow, info->day[2].rain, info->day[2].mintemp, info->day[2].maxtemp, info->day[2].windchill, info->day[2].freezelevel,
				info->day[3].time, info->day[3].summary, info->day[3].wind, info->day[3].snow, info->day[3].rain, info->day[3].mintemp, info->day[3].maxtemp, info->day[3].windchill, info->day[3].freezelevel,
				info->day[4].time, info->day[4].summary, info->day[4].wind, info->day[4].snow, info->day[4].rain, info->day[4].mintemp, info->day[4].maxtemp, info->day[4].windchill, info->day[4].freezelevel,
				info->day[5].time, info->day[5].summary, info->day[5].wind, info->day[5].snow, info->day[5].rain, info->day[5].mintemp, info->day[5].maxtemp, info->day[5].windchill, info->day[5].freezelevel,
				info->day[6].time, info->day[6].summary, info->day[6].wind, info->day[6].snow, info->day[6].rain, info->day[6].mintemp, info->day[6].maxtemp, info->day[6].windchill, info->day[6].freezelevel,
				info->day[7].time, info->day[7].summary, info->day[7].wind, info->day[7].snow, info->day[7].rain, info->day[7].mintemp, info->day[7].maxtemp, info->day[7].windchill, info->day[7].freezelevel,
				info->day[8].time, info->day[8].summary, info->day[8].wind, info->day[8].snow, info->day[8].rain, info->day[8].mintemp, info->day[8].maxtemp, info->day[8].windchill, info->day[8].freezelevel,
				info->day[9].time, info->day[9].summary, info->day[9].wind, info->day[9].snow, info->day[9].rain, info->day[9].mintemp, info->day[9].maxtemp, info->day[9].windchill, info->day[9].freezelevel,
				info->day[10].time, info->day[10].summary, info->day[10].wind, info->day[10].snow, info->day[10].rain, info->day[10].mintemp, info->day[10].maxtemp, info->day[10].windchill, info->day[10].freezelevel,
				info->day[11].time, info->day[11].summary, info->day[11].wind, info->day[11].snow, info->day[11].rain, info->day[11].mintemp, info->day[11].maxtemp, info->day[11].windchill, info->day[11].freezelevel,
				info->day[12].time, info->day[12].summary, info->day[12].wind, info->day[12].snow, info->day[12].rain, info->day[12].mintemp, info->day[12].maxtemp, info->day[12].windchill, info->day[12].freezelevel,
				info->day[13].time, info->day[13].summary, info->day[13].wind, info->day[13].snow, info->day[13].rain, info->day[13].mintemp, info->day[13].maxtemp, info->day[13].windchill, info->day[13].freezelevel,
				info->day[14].time, info->day[14].summary, info->day[14].wind, info->day[14].snow, info->day[14].rain, info->day[14].mintemp, info->day[14].maxtemp, info->day[14].windchill, info->day[14].freezelevel,
				info->day[15].time, info->day[15].summary, info->day[15].wind, info->day[15].snow, info->day[15].rain, info->day[15].mintemp, info->day[15].maxtemp, info->day[15].windchill, info->day[15].freezelevel,
				info->day[16].time, info->day[16].summary, info->day[16].wind, info->day[16].snow, info->day[16].rain, info->day[16].mintemp, info->day[16].maxtemp, info->day[16].windchill, info->day[16].freezelevel,
				info->day[17].time, info->day[17].summary, info->day[17].wind, info->day[17].snow, info->day[17].rain, info->day[17].mintemp, info->day[17].maxtemp, info->day[17].windchill, info->day[17].freezelevel,
				link);
	if (query == NULL) {
		database_close(db);
		return -1;
	}
	debugf("update: %s", query);
	r = sqlite3_exec(db, query, 0, 0, 0);
	sqlite3_free(query);
	database_close(db);
	if (r == SQLITE_OK) {
		return 0;
	} else {
		debugf("error for '%s', with '%s'", link, query);
	}
	return 0;
}

static void * resort_update_thread (void *arg)
{
	int r;
	int p;
	int s;
	int l;
	time_t t;
	char *level;
	char *query;
	char *buffer;
	resort_info_t *info;
	resort_condition_t *condition;
	const char *remote = "http://www.snow-forecast.com/resorts/%s/6day/%s/print";
#if !defined(BUILD_UNIX)
	NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
#endif
	debugf("enter");
	s = 0;
	t = time(NULL);
	condition = (resort_condition_t *) arg;
	pthread_mutex_lock(&condition->mutex);
	condition->started = 1;
	condition->running = 1;
	pthread_cond_signal(&condition->cond);
	pthread_mutex_unlock(&condition->mutex);
	p = resort_update_time(condition->path, condition->resort);
	if (p > (t - DATABE_UPDATE_FREQ)) {
		s = 1;
		goto out;
	}
	debugf("will update:%s @ %d", condition->resort->name, condition->resort->level);
	l = condition->resort->level;
	level = (condition->resort->level == RESORT_LEVEL_TOP) ? "top" :
		(condition->resort->level == RESORT_LEVEL_MID) ? "mid" :
		(condition->resort->level == RESORT_LEVEL_BOT) ? "bot" :
		"mid";
	query = sqlite3_mprintf(remote, condition->resort->link, level);
	if (query == NULL) {
		goto out;
	}
	buffer = buffer_read(query, condition);
	if (buffer == NULL) {
		sqlite3_free(query);
		goto out;
	}
	sqlite3_free(query);
	info = buffer_parse(buffer);
	if (info == NULL) {
		free(buffer);
		goto out;
	}
	buffer_dump(info);
	r = resort_update(condition->path, condition->resort->link, info, l);
	free(buffer);
	free(info->altitude);
	free(info);
	if (r == 0) {
		debugf("success");
		s = 1;
	} else {
		debugf("error for '%s', with '%s'", condition->resort->link, query);
	}
out:	pthread_mutex_lock(&condition->mutex);
	condition->running = 0;
	condition->stopped = 1;
	condition->success = s;
	pthread_cond_signal(&condition->cond);
	pthread_mutex_unlock(&condition->mutex);
#if !defined(BUILD_UNIX)
	[pool release];
#endif
	return NULL;
}

resort_condition_t * resort_update_start (const char *path, resort_t *resort)
{
	int p;
	int t;
	resort_condition_t *condition;
	debugf("starting update thread");
	condition = (resort_condition_t *) malloc(sizeof(resort_condition_t));
	if (condition == NULL) {
		debugf("could not start update thread");
		return NULL;
	}
	memset(condition, 0, sizeof(resort_condition_t));
	if (pthread_mutex_init(&condition->mutex, NULL) != 0) {
		free(condition);
		debugf("could not start update thread");
		return NULL;
	}
	if (pthread_cond_init(&condition->cond, NULL) != 0) {
		pthread_mutex_destroy(&condition->mutex);
		free(condition);
		debugf("could not start update thread");
		return NULL;
	}
	condition->path = strdup(path);
	condition->resort = resort;
	t = time(NULL);
	p = resort_update_time(condition->path, condition->resort);
	if (p > (t - DATABE_UPDATE_FREQ)) {
		condition->started = 1;
		condition->running = 0;
		condition->stopped = 1;
		condition->success = 1;
		goto out;
	}
	if (pthread_create(&condition->thread, NULL, resort_update_thread, condition) != 0) {
		pthread_cond_destroy(&condition->cond);
		pthread_mutex_destroy(&condition->mutex);
		free(condition->path);
		free(condition);
		debugf("could not start update thread");
		return NULL;
	}
	pthread_mutex_lock(&condition->mutex);
	while (condition->started == 0) {
		pthread_cond_wait(&condition->cond, &condition->mutex);
	}
	pthread_mutex_unlock(&condition->mutex);
	debugf("started update thread");
out:	return condition;
}

int resort_update_check (resort_condition_t *condition)
{
	int ret;
	ret = -1;
	if (condition == NULL) {
		return -3;
	}
	pthread_mutex_lock(&condition->mutex);
	if (condition->stopped == 1) {
		if (condition->success == 1) {
			ret = 0;
		} else {
			ret = -2;
		}
	}
	pthread_mutex_unlock(&condition->mutex);
	return ret;
}

int resort_update_finish (resort_condition_t *condition)
{
	if (condition == NULL) {
		return 0;
	}
	debugf("stopping update tread");
	pthread_mutex_lock(&condition->mutex);
	condition->running = 0;
	pthread_cond_signal(&condition->cond);
	while (condition->stopped == 0) {
		pthread_cond_wait(&condition->cond, &condition->mutex);
	}
	pthread_join(condition->thread, NULL);
	pthread_mutex_unlock(&condition->mutex);
	pthread_cond_destroy(&condition->cond);
	pthread_mutex_destroy(&condition->mutex);
	free(condition->path);
	free(condition);
	debugf("stopped update tread");
	return 0;
}

resort_t ** resorts_list (void)
{
	return (resort_t **) resorts_table;
}

static inline int resort_free (resort_t *r)
{
	int i;
	debugf("enter");
	free(r->country);
	free(r->link);
	free(r->name);
	free(r->info.altitude);
	free(r->info.header);
	free(r->info.info);
	free(r->info.summary[0]);
	free(r->info.summary[1]);
	for (i = 0; i < 18; i++) {
		free(r->info.day[i].freezelevel);
		free(r->info.day[i].maxtemp);
		free(r->info.day[i].mintemp);
		free(r->info.day[i].rain);
		free(r->info.day[i].snow);
		free(r->info.day[i].summary);
		free(r->info.day[i].sunrise);
		free(r->info.day[i].sunset);
		free(r->info.day[i].time);
		free(r->info.day[i].weather);
		free(r->info.day[i].wind);
		free(r->info.day[i].windchill);
	}
	free(r);
	debugf("leave");
	return 0;
}

int resorts_free (resort_t **resorts)
{
	resort_t **r;
	if (resorts == (resort_t **) resorts_table) {
		return 0;
	}
	for (r = resorts; r && *r; r++) {
		resort_free(*r);
	}
	free(resorts);
	return 0;
}

int resorts_count (resort_t **resorts)
{
	int c;
	resort_t **r;
	debugf("enter");
	c = 0;
	if (resorts == (resort_t **) resorts_table) {
		c = sizeof(resorts_table) / sizeof(resorts_table[0]);
	} else {
		for (r = resorts; r && *r; r++) {
			c++;
		}
	}
	debugf("leave (c=%d)", c);
	return c;
}

resort_t * resorts_resort (resort_t **resorts, int index)
{
	if (resorts) {
		return resorts[index];
	}
	return NULL;
}

int resorts_setlevel (const char *path, resort_level_t level)
{
	int r;
	time_t t;
	sqlite3 *db;
	char *query;
	const char *format = "UPDATE RESORT set TIME='%u', TMB='%u';";
	t = 0;
	db = database_open(path);
	if (db == NULL) {
		return -1;
	}
	query = sqlite3_mprintf(format, t, level);
	if (query == NULL) {
		database_close(db);
		return -1;
	}
	r = sqlite3_exec(db, query, 0, 0, 0);
	sqlite3_free(query);
	database_close(db);
	if (r == SQLITE_OK) {
		return 0;
	} else {
		debugf("error for '%s', with '%s'", link, query);
	}
	return 0;
}

resort_t ** resorts_read (const char *path)
{
	int c;
	int e;
	int i;
	char *t;
	resort_t **r;
	resort_t **rs;
	resort_t **rt;
	sqlite3_stmt *stmt;
	const char *query =
		"SELECT"
		"  r.NAME, r.COUNTRY, r.LINK, r.TMB,"
		"  r.ALTITUDE, r.INFO, r.SUMMARY0, r.SUMMARY1,"
		"  r.DAY0TIME, r.DAY0SUMMARY, r.DAY0WIND, r.DAY0SNOW, r.DAY0RAIN, r.DAY0TEMPMIN, r.DAY0TEMPMAX, r.DAY0TEMPCHILL, r.DAY0FREEZE,"
		"  r.DAY1TIME, r.DAY1SUMMARY, r.DAY1WIND, r.DAY1SNOW, r.DAY1RAIN, r.DAY1TEMPMIN, r.DAY1TEMPMAX, r.DAY1TEMPCHILL, r.DAY1FREEZE,"
		"  r.DAY2TIME, r.DAY2SUMMARY, r.DAY2WIND, r.DAY2SNOW, r.DAY2RAIN, r.DAY2TEMPMIN, r.DAY2TEMPMAX, r.DAY2TEMPCHILL, r.DAY2FREEZE,"
		"  r.DAY3TIME, r.DAY3SUMMARY, r.DAY3WIND, r.DAY3SNOW, r.DAY3RAIN, r.DAY3TEMPMIN, r.DAY3TEMPMAX, r.DAY3TEMPCHILL, r.DAY3FREEZE,"
		"  r.DAY4TIME, r.DAY4SUMMARY, r.DAY4WIND, r.DAY4SNOW, r.DAY4RAIN, r.DAY4TEMPMIN, r.DAY4TEMPMAX, r.DAY4TEMPCHILL, r.DAY4FREEZE,"
		"  r.DAY5TIME, r.DAY5SUMMARY, r.DAY5WIND, r.DAY5SNOW, r.DAY5RAIN, r.DAY5TEMPMIN, r.DAY5TEMPMAX, r.DAY5TEMPCHILL, r.DAY5FREEZE,"
		"  r.DAY6TIME, r.DAY6SUMMARY, r.DAY6WIND, r.DAY6SNOW, r.DAY6RAIN, r.DAY6TEMPMIN, r.DAY6TEMPMAX, r.DAY6TEMPCHILL, r.DAY6FREEZE,"
		"  r.DAY7TIME, r.DAY7SUMMARY, r.DAY7WIND, r.DAY7SNOW, r.DAY7RAIN, r.DAY7TEMPMIN, r.DAY7TEMPMAX, r.DAY7TEMPCHILL, r.DAY7FREEZE,"
		"  r.DAY8TIME, r.DAY8SUMMARY, r.DAY8WIND, r.DAY8SNOW, r.DAY8RAIN, r.DAY8TEMPMIN, r.DAY8TEMPMAX, r.DAY8TEMPCHILL, r.DAY8FREEZE,"
		"  r.DAY9TIME, r.DAY9SUMMARY, r.DAY9WIND, r.DAY9SNOW, r.DAY9RAIN, r.DAY9TEMPMIN, r.DAY9TEMPMAX, r.DAY9TEMPCHILL, r.DAY9FREEZE,"
		"  r.DAY10TIME, r.DAY10SUMMARY, r.DAY10WIND, r.DAY10SNOW, r.DAY10RAIN, r.DAY10TEMPMIN, r.DAY10TEMPMAX, r.DAY10TEMPCHILL, r.DAY10FREEZE,"
		"  r.DAY11TIME, r.DAY11SUMMARY, r.DAY11WIND, r.DAY11SNOW, r.DAY11RAIN, r.DAY11TEMPMIN, r.DAY11TEMPMAX, r.DAY11TEMPCHILL, r.DAY11FREEZE,"
		"  r.DAY12TIME, r.DAY12SUMMARY, r.DAY12WIND, r.DAY12SNOW, r.DAY12RAIN, r.DAY12TEMPMIN, r.DAY12TEMPMAX, r.DAY12TEMPCHILL, r.DAY12FREEZE,"
		"  r.DAY13TIME, r.DAY13SUMMARY, r.DAY13WIND, r.DAY13SNOW, r.DAY13RAIN, r.DAY13TEMPMIN, r.DAY13TEMPMAX, r.DAY13TEMPCHILL, r.DAY13FREEZE,"
		"  r.DAY14TIME, r.DAY14SUMMARY, r.DAY14WIND, r.DAY14SNOW, r.DAY14RAIN, r.DAY14TEMPMIN, r.DAY14TEMPMAX, r.DAY14TEMPCHILL, r.DAY14FREEZE,"
		"  r.DAY15TIME, r.DAY15SUMMARY, r.DAY15WIND, r.DAY15SNOW, r.DAY15RAIN, r.DAY15TEMPMIN, r.DAY15TEMPMAX, r.DAY15TEMPCHILL, r.DAY15FREEZE,"
		"  r.DAY16TIME, r.DAY16SUMMARY, r.DAY16WIND, r.DAY16SNOW, r.DAY16RAIN, r.DAY16TEMPMIN, r.DAY16TEMPMAX, r.DAY16TEMPCHILL, r.DAY16FREEZE,"
		"  r.DAY17TIME, r.DAY17SUMMARY, r.DAY17WIND, r.DAY17SNOW, r.DAY17RAIN, r.DAY17TEMPMIN, r.DAY17TEMPMAX, r.DAY17TEMPCHILL, r.DAY17FREEZE"
		"  from RESORT r;";
	sqlite3 *db;
	debugf("enter");
	db = database_open(path);
	if (db == NULL) {
		debugf("error");
		return NULL;
	}
	e = 0;
	c = 0;
	rs = NULL;
	stmt = NULL;
	#define sm() { \
		c++; \
		rt = rs; \
		rs = (resort_t **) malloc(sizeof(resort_t *) * (c + 1)); \
		if (rs == NULL) { \
			e = 1; \
			break; \
		} \
		memset(rs, 0, sizeof(resort_t *) * (c + 1)); \
		if (c > 1) { \
			memcpy(rs, rt, sizeof(resort_t *) * (c - 1)); \
		} \
		free(rt); \
		r = rs + (c - 1); \
		*r = (resort_t *) malloc(sizeof(resort_t)); \
		if (*r == NULL) { \
			e = 1; \
			break; \
		} \
		memset(*r, 0, sizeof(resort_t)); \
	}
	#define si(n, a) { \
		t = (char *) sqlite3_column_text(stmt, (n)); \
		(*r)->a = strdup(t); \
		if ((*r)->a == NULL) { \
			e = 1; \
			break; \
		} \
	}
	#define sc(n, a) { \
		t = (char *) sqlite3_column_text(stmt, (n)); \
		(*r)->a = atoi(t); \
	}
	if (sqlite3_prepare_v2(db, query, -1, &stmt, NULL) == SQLITE_OK) {
		while (sqlite3_step(stmt) == SQLITE_ROW) {
			sm();
			si(0, name);
			si(1, country);
			si(2, link);
			sc(3, level);
			si(4, info.altitude);
			si(5, info.info);
			si(6, info.summary[0]);
			si(7, info.summary[1]);
			for (i = 0; i < 18; i++) {
				si(8 + (i * 9) + 0, info.day[i].time);
				si(8 + (i * 9) + 1, info.day[i].summary);
				si(8 + (i * 9) + 2, info.day[i].wind);
				si(8 + (i * 9) + 3, info.day[i].snow);
				si(8 + (i * 9) + 4, info.day[i].rain);
				si(8 + (i * 9) + 5, info.day[i].mintemp);
				si(8 + (i * 9) + 6, info.day[i].maxtemp);
				si(8 + (i * 9) + 7, info.day[i].windchill);
				si(8 + (i * 9) + 8, info.day[i].freezelevel);
			}
			r++;
		}
	}
	sqlite3_finalize(stmt);
	database_close(db);
	if (e != 0) {
		resorts_free(rs);
		rs = NULL;
	}
	debugf("leave (rs:%p) (count:%d)", rs, resorts_count(rs));
	return rs;
}

#if defined(BUILD_UNIX)
int main (int argc, char *argv[])
{
	char *buffer;
	resort_info_t *data;
	if (argc != 2) {
		printf("usage: %s input\n", argv[0]);
		return -1;
	}
	buffer = buffer_read(argv[1], NULL);
	if (buffer == NULL) {
		printf("buffer_read('%s') failed\n", argv[1]);
		return -2;
	}
	data = buffer_parse(buffer);
	buffer_dump(data);
	free(buffer);
	free(data->altitude);
	free(data);
	return 0;
}
#endif
