
#import <QuartzCore/QuartzCore.h> /* for cornerRadius */
#import "SnowForeCastAppDelegate.h"
#import "SwitchViewController.h"
#import "SettingsViewController.h"
#import "Logger.h"
#import "resort.h"

@implementation SettingsViewController

@synthesize tableView;
@synthesize infoSegment;
@synthesize searchBar;
@synthesize toolBar;
@synthesize isSearching;
@synthesize searchString;
@synthesize doneButton;
@synthesize switchView;
@synthesize toolLabel;

- (void) viewDidLoad
{
	SnowForeCastAppDelegate *appDelegate;
	appDelegate = (SnowForeCastAppDelegate *) [[UIApplication sharedApplication] delegate];
	switchView = appDelegate.switchViewController;
	[self updateUI];
	[super viewDidLoad];
}

- (void) didReceiveMemoryWarning
{
	logf(@"memory warning");
	[super didReceiveMemoryWarning];
}

- (void) viewDidUnload
{
	logf(@"unloading");
}

- (void) dealloc {
	logf(@"releasing");
	[searchString release];
	[super dealloc];
}

- (NSInteger) tableView:(UITableView *) tableView numberOfRowsInSection:(NSInteger) section
{
	unsigned int c;
	resort_t **resort;
	resort_t **resorts;
	if (isSearching == YES) {
		resorts = resorts_list();
		if (searchString == nil) {
			c = resorts_count(resorts);
			logf(@"resorts count:%d", c);
			return c;
		} else {
			c = 0;
			for (resort = resorts; *resort; resort++) {
				if (strcasestr((*resort)->link, [searchString UTF8String]) != NULL ||
					strcasestr((*resort)->country, [searchString UTF8String]) != NULL) {
					c++;
				}
			}
			logf(@"resorts count:%d", c);
			return c;
		}
	} else {
		c = resorts_count(switchView.resortsArray);
		logf(@"resorts count:%d", c);
		return c;
	}
}

- (UITableViewCell *) tableView:(UITableView *) sender cellForRowAtIndexPath:(NSIndexPath *) indexPath
{
	unsigned int c;
	resort_t *res;
	resort_t **resort;
	resort_t **resorts;
	UITableViewCell *cell;
	static NSString *cellIdentifier = @"Cell";
	cell = [tableView dequeueReusableCellWithIdentifier:cellIdentifier];
	if (cell == nil) {
		cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:cellIdentifier] autorelease];
	}
	if (isSearching == YES) {
		resorts = resorts_list();
		if (searchString == nil) {
			cell.textLabel.text = [NSString stringWithFormat:@"%s, %s", resorts[indexPath.row]->link, resorts[indexPath.row]->country];
		} else {
			c = 0;
			for (resort = resorts; *resort; resort++) {
				if (strcasestr((*resort)->link, [searchString UTF8String]) != NULL ||
					strcasestr((*resort)->country, [searchString UTF8String]) != NULL) {
					if (c == indexPath.row) {
						cell.textLabel.text = [NSString stringWithFormat:@"%s, %s", (*resort)->link, (*resort)->country];
						logf(@"cell[%04d]: %@", indexPath.row, cell.textLabel.text);
						break;
					}
					c++;
				}
			}
		}
	} else {
		res = resorts_resort(switchView.resortsArray, indexPath.row);
		if (res != NULL) {
			cell.textLabel.text = [NSString stringWithFormat:@"%s, %s", res->link, res->country];
		}
		logf(@"cell[%04d]: %@", indexPath.row, cell.textLabel.text);
	}
	return cell;
}

- (BOOL) tableView:(UITableView *) tableView canMoveRowAtIndexPath:(NSIndexPath *) indexPath
{
	return NO;
}

- (UITableViewCellEditingStyle) tableView:(UITableView *) tableView editingStyleForRowAtIndexPath:(NSIndexPath *) indexPath
{
	if (isSearching == YES) {
		return UITableViewCellEditingStyleNone;
	} else {
		return UITableViewCellEditingStyleDelete;
	}
}

- (void) tableView:(UITableView *) tableView moveRowAtIndexPath:(NSIndexPath *) sourceIndexPath toIndexPath:(NSIndexPath *) destinationIndexPath;
{
	logf(@"moving cell %d -> %d", sourceIndexPath.row, destinationIndexPath.row);
}

- (void) tableView:(UITableView *) tableView commitEditingStyle:(UITableViewCellEditingStyle) editingStyle forRowAtIndexPath:(NSIndexPath *) indexPath
{
	int ret;
	resort_t *resort;
	if (editingStyle == UITableViewCellEditingStyleDelete) {
		resort = resorts_resort(switchView.resortsArray, indexPath.row);
		if (resort != nil) {
			logf(@"selected resort: %s, %s (- deleting)\n", resort->link, resort->country);
			ret = resort_remove([switchView.databasePath UTF8String], resort);
			resorts_free(switchView.resortsArray);
			switchView.resortsArray = resorts_read([switchView.databasePath UTF8String]);
			if (ret == 0) {
				[self.tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath] withRowAnimation:UITableViewRowAnimationFade];
			}
			if (resorts_count(switchView.resortsArray) <= 0) {
				[doneButton setEnabled:NO];
			} else {
				[doneButton setEnabled:YES];
			}
		}
	}
}

- (void) tableView:(UITableView *) sender didSelectRowAtIndexPath:(NSIndexPath *) indexPath
{
	int ret;
	unsigned int c;
	resort_t **resort;
	resort_t **resorts;
	if (isSearching == NO) {
		[tableView deselectRowAtIndexPath:indexPath animated:YES];
		return;
	}
	c = 0;
	resorts = resorts_list();
	for (resort = resorts; *resort; resort++) {
		if (searchString != nil) {
			if (strcasestr((*resort)->link, [searchString UTF8String]) != NULL ||
				strcasestr((*resort)->country, [searchString UTF8String]) != NULL) {
				if (c == indexPath.row) {
					break;
				}
				c++;
			}
		} else {
			if (c == indexPath.row) {
				break;
			}
			c++;
		}
	}
	ret = -1;
	if (*resort) {
		int level;
		logf(@"selected resort: %s, %s (+ inserting)\n", (*resort)->link, (*resort)->country);
		if (resorts_count(switchView.resortsArray) <= 0) {
			level = RESORT_LEVEL_DEFAULT;
		} else {
			level = switchView.resortsArray[0]->level;
		}
		(*resort)->level = level;
		ret = resort_insert([switchView.databasePath UTF8String], *resort);
	}
	[tableView deselectRowAtIndexPath:indexPath animated:YES];
	if (ret == 0) {
		resorts_free(switchView.resortsArray);
		switchView.resortsArray = resorts_read([switchView.databasePath UTF8String]);
		[self stopSearch];
//		[switchView switchViews:self];
	}
}

- (void) searchBar:(UISearchBar *) sender textDidChange:(NSString *) searchText;
{
	logf(@"searching:%@", searchText);
	if (searchString != nil) {
		[searchString release];
		searchString = nil;
	}
	if (searchText == nil ||
		searchText == @"") {
		return;
	}
	searchString = [[NSString alloc] initWithString:searchText];
	[tableView reloadData];
}

- (void) searchBarCancelButtonClicked:(UISearchBar *) sender
{
	[searchString release];
	searchString = nil;
	[self stopSearch];
}

- (IBAction) infoChanged:(id) sender
{
	NSInteger index;
	index = [[self infoSegment] selectedSegmentIndex];
	logf(@"cell info changed to %d", index);
	resorts_setlevel([switchView.databasePath UTF8String],
			 (index == 0) ? RESORT_LEVEL_TOP :
			 (index == 1) ? RESORT_LEVEL_MID :
			 (index == 2) ? RESORT_LEVEL_BOT :
			 RESORT_LEVEL_DEFAULT);
	resorts_free(switchView.resortsArray);
	switchView.resortsArray = resorts_read([switchView.databasePath UTF8String]);
}

- (IBAction) startSearch:(id) sender
{
	logf(@"starting search");
	isSearching = YES;
	[searchBar setHidden:NO];
	[toolBar setHidden:YES];
	[toolLabel setHidden:YES];
	[tableView setEditing:NO animated:YES];
	[tableView reloadData];
	[tableView scrollToRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0] atScrollPosition:UITableViewScrollPositionNone animated:NO];
	[searchBar becomeFirstResponder];
}

- (void) stopSearch
{
	logf(@"stopping search");
	isSearching = NO;
	searchBar.text = @"";
	[searchBar resignFirstResponder];
	[toolBar setHidden:NO];
	[toolLabel setHidden:NO];
	[searchBar setHidden:YES];
	[tableView setEditing:YES animated:YES];
	[tableView reloadData];
	if (resorts_count(switchView.resortsArray) <= 0) {
		[doneButton setEnabled:NO];
		infoSegment.selectedSegmentIndex = 1;
		[infoSegment setEnabled:NO];
	} else {
		[doneButton setEnabled:YES];
		[infoSegment setEnabled:YES];
		switch (switchView.resortsArray[0]->level) {
			case RESORT_LEVEL_TOP:
				infoSegment.selectedSegmentIndex = 0;
				break;
			default:
			case RESORT_LEVEL_MID:
				infoSegment.selectedSegmentIndex = 1;
				break;
			case RESORT_LEVEL_BOT:
				infoSegment.selectedSegmentIndex = 2;
				break;
		}
		[tableView scrollToRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0] atScrollPosition:UITableViewScrollPositionNone animated:NO];
	}
}

- (void) updateUI
{
	logf(@"updating ui");
	tableView.layer.cornerRadius = 15;
	tableView.allowsSelectionDuringEditing = NO;
	[searchString release];
	searchString = nil;
	[self stopSearch];
}

@end
