
#import "SnowForeCastAppDelegate.h"
#import "SwitchViewController.h"
#import "SnowViewController.h"
#import "SettingsViewController.h"

#import "Resort.h"
#import "Logger.h"

@implementation SwitchViewController

@synthesize snowViewController;
@synthesize settingsViewController;
@synthesize databasePath;
@synthesize resortsArray;

- (void) viewDidLoad
{
	NSInteger count;

	NSArray *documentPaths;
	NSString *documentsDir;
	
	NSString *databaseName;
	SnowViewController *snowController;
	SettingsViewController *settingsController;
	
	databaseName = @"SnowForecast.sql";
	documentPaths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
	documentsDir = [documentPaths objectAtIndex:0];
	databasePath = [[NSString alloc] initWithString:[documentsDir stringByAppendingPathComponent:databaseName]];
	logf(@"switchView setting databasePath:'%@'", databasePath);
	
	resortsArray = resorts_read([databasePath UTF8String]);
	count = resorts_count(resortsArray);
	
	if (count <= 0) {
		logf(@"switchView loading settingsView (count:%d)", count);
		settingsController = [[SettingsViewController alloc] initWithNibName:@"SettingsView" bundle:nil];
		self.settingsViewController = settingsController;
		[self.view insertSubview:settingsController.view atIndex:0];
		[settingsController release];
	} else {
		logf(@"switchView loading snowView (count:%d)", count);
		snowController = [[SnowViewController alloc] initWithNibName:@"SnowView" bundle:nil];
		self.snowViewController = snowController;
		[self.view insertSubview:snowController.view atIndex:0];
		[snowController release];
	}
    [super viewDidLoad];
}

- (IBAction) switchViews: (id) sender
{
	[UIView beginAnimations:@"View Flip" context:nil];
	[UIView setAnimationDuration:0.75];
	[UIView setAnimationCurve:UIViewAnimationCurveEaseInOut];
	if (self.settingsViewController.view.superview == nil) {
		logf(@"switchView switching to settingsView");
		if (self.settingsViewController == nil) {
			SettingsViewController *settingsController;
			settingsController = [[SettingsViewController alloc] initWithNibName:@"SettingsView" bundle:nil];
			self.settingsViewController = settingsController;
			[settingsController release];
		} else {
			[settingsViewController updateUI];
		}
		[UIView setAnimationTransition:UIViewAnimationTransitionFlipFromRight forView:self.view cache:YES];
		[snowViewController viewWillAppear:YES];
		[settingsViewController viewWillDisappear:YES];
		[snowViewController.view removeFromSuperview];
		[self.view insertSubview:settingsViewController.view atIndex:0];
		[settingsViewController viewDidDisappear:YES];
		[snowViewController viewDidAppear:YES];
	} else {
		logf(@"switchView switching to snowView");
		if (self.snowViewController == nil) {
			SnowViewController *snowController;
			snowController = [[SnowViewController alloc] initWithNibName:@"SnowView" bundle:nil];
			self.snowViewController = snowController;
			[snowController release];
		} else {
			[snowViewController updateUI];
		}
		[UIView setAnimationTransition:UIViewAnimationTransitionFlipFromLeft forView:self.view cache:YES];
		[settingsViewController viewWillAppear:YES];
		[snowViewController viewWillDisappear:YES];
		[settingsViewController.view removeFromSuperview];
		[self.view insertSubview:snowViewController.view atIndex:0];
		[snowViewController viewDidDisappear:YES];
		[settingsViewController viewDidAppear:YES];
	}
	[UIView commitAnimations];
}

- (void) didReceiveMemoryWarning {
    [super didReceiveMemoryWarning];
	if (self.snowViewController.view.superview == nil) {
		self.snowViewController = nil;
	} else {
		self.settingsViewController = nil;
	}
}

- (void) dealloc
{
	[snowViewController release];
	[settingsViewController release];
	[databasePath release];
    [super dealloc];
}

@end
